import os

import yaml

from clifire import out


class Config:
    @classmethod
    def get_config(cls, files: list, create: bool = False, **kwargs):
        def path(*args) -> str:
            exapnd_path = os.path.join(
                *(a.replace("~", os.path.expanduser("~")) for a in args)
            )
            return os.path.abspath(exapnd_path)

        for file in files:
            config_file = path(file)
            out.debug(f"Try read config file {config_file}")
            if os.path.exists(config_file):
                out.debug2("Exists, reading...")
                config = Config(config_file=config_file, **kwargs)
                config.read()
                return config
            out.debug2("Not exist")
        if files:
            file = path(files[0])
            config = Config(config_file=config_file, **kwargs)
            if create:
                config.write()
            return config

    def __init__(self, config_file, **kwargs):
        self._config_file = config_file
        self._config_path = os.path.dirname(config_file)
        self.__dict__.update(kwargs)

    def get(self, key, default=None):
        return getattr(self, key, default)

    def __str__(self) -> str:
        return str(self.__dict__)

    def __repr__(self) -> str:
        return self.__str__()

    def _safe_dict(self, data_dict: dict):
        return {k: v for k, v in data_dict.items() if not k.startswith("_")}

    def read(self):
        if not os.path.exists(self._config_file):
            return False
        with open(self._config_file, "r", encoding="utf-8") as file:
            self.__dict__.update(self._safe_dict(yaml.safe_load(file) or {}))
        return True

    def write(self):
        out.debug(f"Write config file {self._config_file}")
        config_dir = os.path.dirname(self._config_file)
        if not os.path.exists(config_dir):
            os.makedirs(config_dir)
        with open(self._config_file, "w", encoding="utf-8") as file:
            yaml.safe_dump(self._safe_dict(self.__dict__), file)



import os
import shlex
import subprocess
import sys

from clifire import command, commands, config, out, result, template


class App:
    current_app = None

    def __init__(
        self,
        name: str = "",
        version: str = "0.0.1 alpha",
        context: dict = None,
        option_verbose: bool = True,
        config_files: list = None,
        config_create: bool = False,
        command_help=commands.help.CommandHelp,
        command_version=commands.version.CommandVersion,
        template_folder=None,
    ):
        App.current_app = self
        self.name = name
        self.version = version
        self.options = {}
        if option_verbose:
            self._add_option_verbose()
        if context is None:
            context = {}
        self.context = context
        if config_files is None:
            config_files = []
        self.config = config.Config.get_config(
            config_files, create=config_create, **context
        )
        self.commands = {}
        if command_help:
            self.add_command(command_help)
        if command_version:
            self.add_command(command_version)
        self.template = None
        if template_folder:
            self.template = template.Template(template_folder)

    def _add_option_verbose(self):
        self.add_option(
            "verbose",
            command.Field(
                help="Verbose mode",
                default=False,
                alias="v",
            ),
        )
        command_line = sys.argv[1:]
        if "-v" in command_line or "--verbose" in command_line:
            self.set_option("verbose", True)

    def add_option(self, name: str, field: command.Field):
        self.options[name] = [field, field.default]
        for alias in field.alias:
            if alias.startswith("-"):
                alias = alias[2:] if alias.startswith("--") else alias[1:]
            alias = alias.replace("-", "_")
            if alias in self.options:
                raise command.CommandException(
                    f'Duplicate global option alias "{alias}"'
                )
            self.options[alias] = [name, None]

    def set_option(self, name: str, value):
        if name not in self.options:
            return False
        self.options[name][1] = value
        return True

    def get_option(self, name: str, default=None):
        if name not in self.options:
            return default
        return self.options[name][1]

    def add_command(self, cls: command.Command):
        if not cls._name:
            raise command.CommandException(
                f"The command {cls} has no name, please set _name var in class"
            )
        self.commands[cls._name] = cls

    def find_command(self, command_line: str):
        empty = False
        args = [p for p in shlex.split(command_line) if not p.startswith("-")]
        if not args:
            args = ["help"]
            empty = True
        while args:
            command_name = ".".join(args)
            if command_name in self.commands:
                return self.commands[command_name]
            args.pop()
        if empty:
            out.critical("No command provided.", code=10)
        return None

    def get_command(self, command_line: str) -> command.Command:
        cls = self.find_command(command_line)
        if cls:
            return cls(self, command_line)
        args = [p for p in shlex.split(command_line) if not p.startswith("-")]
        out.critical(f'Command "{args[0]}" not found.', code=20)

    def fire(self, command_line: str = None):
        try:
            if command_line is None:
                out.debug(f"Sys argv value: {sys.argv}")
                command_line = " ".join(
                    shlex.quote(arg) for arg in sys.argv[1:]
                )
            cmd = self.get_command(command_line)
            res = cmd.launch(command_line)
            if type(res) is int and res != 0:
                sys.exit(res)
        except command.CommandException as e:
            out.critical(e, code=30)
        except command.FieldException as e:
            out.critical(e, code=40)

    def shell(
        self,
        cmd: str,
        capture_output: bool = True,
        env: dict = None,
        path: str = None,
        shell: bool = True,
        check: bool = False,
    ) -> result.Result:
        if path:
            os.chdir(path)
        env_vars = os.environ.copy()
        if env:
            env_vars.update(env)
        try:
            out.debug(f"Shell: {cmd}")
            proc = subprocess.run(
                cmd if shell else shlex.split(cmd),
                shell=shell,
                capture_output=capture_output,
                env=env_vars,
                check=check,
            )
            return result.Result(proc.returncode, proc.stdout, proc.stderr)
        except subprocess.CalledProcessError as e:
            return result.ResultError(e.stderr, e.returncode)

    def path(self, *args) -> str:
        exapnd_path = os.path.join(
            *(a.replace("~", os.path.expanduser("~")) for a in args)
        )
        return os.path.abspath(exapnd_path)



import inspect
import re
import shlex
from typing import List, Optional, Union

from clifire import out


def get_current_app():
    from clifire import application

    return application.App.current_app


def fire(func):
    def wrapper(*args, **kwargs):
        cmd = args[0]
        signature = inspect.signature(func)
        for arg in list(signature.parameters.keys())[1:]:
            var = arg[1:] if arg.startswith("_") else arg
            if hasattr(cmd, var):
                kwargs[arg] = getattr(cmd, var)
        return func(*args, **kwargs)

    command_name = getattr(func, "_command_name", func.__name__)
    command_name = re.sub(r"([A-Z_])", ".", command_name).lower()
    class_name = "".join(
        word.capitalize() for word in command_name.split(".") if word
    )
    doc = (func.__doc__ or "").strip()
    attrs = {
        "_name": command_name,
        "_help": doc.splitlines()[0] if doc.splitlines() else doc,
        "fire": wrapper,
    }
    signature = inspect.signature(func)
    pos = 0
    helps = {k: [] for k in list(signature.parameters.keys())[1:]}
    doc = [d.strip() for d in doc.splitlines() if d]
    current_help = False
    while doc:
        line = doc.pop(0)
        parts = line.split(":")
        if parts[0] in helps:
            current_help = parts[0]
            line = parts[1]
        if not current_help:
            continue
        helps[current_help].append(line.strip())
    for name, param in list(signature.parameters.items())[1:]:
        pos += 1
        var_name = name[1:] if name.startswith("_") else name
        attrs[var_name] = Field(
            pos=None if name.startswith("_") else pos,
            help=" ".join(helps.get(name, "")),
            default=param._default,
            force_type=param._annotation,
            alias=[var_name[0]],
        )
    class_command = type(class_name, (Command,), attrs)
    app = get_current_app()
    app.add_command(class_command)
    return wrapper


class Field:
    def __init__(
        self,
        pos: int = False,
        help: str = "",
        default: str = None,
        alias: Optional[Union[str, List[str]]] = None,
        force_type: type = None,
    ):
        self.name = "unknow"
        self.pos = pos
        self.help = help
        alias = [] if alias is None else alias
        self.alias = [alias] if isinstance(alias, str) else alias
        self.default = default
        self.is_option = bool(pos is False or pos is None)
        self.is_required = default is None
        if force_type is None:
            if self.default is not None:
                force_type = type(self.default)
            elif self.is_option:
                force_type = bool
            else:
                force_type = str
        self.type = force_type

    def convert(self, value):
        try:
            if self.type == list:
                return value.split(",")
            elif self.type == bool and value is None:
                return True if self.default is None else not self.default
            return self.type(value)
        except Exception as exc:
            msg = f'with he value "{value}" must be {self.type.__name__}'
            raise FieldException(self, msg) from exc


class FieldException(Exception):
    def __init__(self, field: Field, msg: str):
        self.field = field
        field_type = "option" if self.field.is_option else "argument"
        super().__init__(f'The {field_type} "{field.name}" {msg}')


class Command:
    _name = ""
    _help = None

    def __init__(self, app, command_line: str = ""):
        self._fields = {}
        self._argument_names = []
        self._options = {}
        self.app = app
        self._fields_update()
        self.command_line = ""

    @property
    def context(self):
        return self.app.context

    def _fields_update(self):
        for name in dir(self):
            field = getattr(self, name)
            if isinstance(field, Field):
                field.name = name
                self._fields[name] = field
        self._argument_names = sorted(
            [key for key, val in self._fields.items() if not val.is_option],
            key=lambda k: self._fields[k].pos,
        )
        self._options = {}
        for name, field in self._fields.items():
            if not field.is_option:
                continue
            self._options[name] = field
            for alias in field.alias:
                if alias.startswith("-"):
                    alias = alias[2:] if alias.startswith("--") else alias[1:]
                alias = alias.replace("-", "_")
                if alias in self._options:
                    raise CommandException(f'Duplicate option alias "{alias}"')
                self._options[alias] = name

    def _fields_check(self):
        for name, field in self._fields.items():
            value = getattr(self, name)
            if isinstance(value, Field):
                if field.is_required:
                    raise FieldException(field, "is required")
                setattr(self, name, field.default)

    def _parse_command_line(self, command_line: str):
        out.debug(f"Parse command line: {command_line}")
        arguments = []
        self.command_line = shlex.split(command_line)
        parts = self.command_line.copy()
        remove_parts = len(self._name.split("."))
        while parts:
            part = parts.pop(0)
            if not part.startswith("-"):
                remove_parts -= 1
                if remove_parts < 0:
                    arguments.append(part)
                continue
            option = part[2:] if part.startswith("--") else part[1:]
            name, value = (
                option.split("=", 1) if "=" in option else (option, None)
            )
            name = name.replace("-", "_")
            if name not in self._options:
                if name not in self.app.options:
                    continue
                field, _value = self.app.options[name]
                if isinstance(field, str):
                    name = field
                    field, _value = self.app.options[name]
                if not value and field.type != bool:
                    value = parts.pop(0)
                value = field.convert(value)
                self.app.set_option(name, value)
                out.debug2(f'Global option "{name}" = {value}')
                continue
            field = self._options[name]
            if isinstance(field, str):
                name = field
                field = self._options[name]
            if not value and field.type != bool:
                value = parts.pop(0)
            value = field.convert(value)
            out.debug2(f'Option "{name}" = {value}')
            setattr(self, name, value)
        arg_names = self._argument_names.copy()
        for index, argument in enumerate(arguments):
            if not arg_names:
                break
            name = arg_names.pop(0)
            if self._fields[name].type == list:
                out.debug2(f'Argument "{name}" = {arguments[index:]}')
                setattr(self, name, arguments[index:])
                break
            value = self._fields[name].convert(argument)
            out.debug2(f'Argument "{name}" = {value}')
            setattr(self, name, value)

    def parse(self, command_line: str):
        self._parse_command_line(command_line)
        self._fields_check()

    def launch(self, command_line: str):
        out.debug(f"Launching command '{self._name}'")
        self.parse(command_line)
        out.debug(f"Init command '{self._name}'")
        self.init()
        out.debug(f"Running command '{self._name}'")
        return self.fire()

    def init(self):
        pass

    def fire(self):
        raise NotImplementedError


class CommandException(Exception):
    pass



import os
import re

import jinja2


class Template:
    def __init__(self, template_folder: str):
        self.jinja2 = jinja2.Environment(
            loader=jinja2.FileSystemLoader(template_folder)
        )

    def render(self, template, **args):
        args["os"] = os
        return self.jinja2.get_template(template).render(**args)

    def write(self, template, filename, mark=None, **args):
        template_content = self.render(template, **args)

        if not mark:
            with open(filename, "w") as file:
                file.write(template_content)
            return template_content

        content = ""
        if os.path.exists(filename):
            with open(filename, "r") as file:
                content = file.read()
        mark_escaped = re.escape(mark)
        pattern = f"({mark_escaped})[\\s\\S]*?\\1"
        if re.search(pattern, content):
            content = re.sub(
                pattern, f"{mark}\n{template_content}\n{mark}", content
            )
        else:
            content += f"\n{mark}\n{template_content}\n{mark}\n"
        with open(filename, "w") as file:
            file.write(content)
        return content



import re
import sys
import threading
import time
from typing import Any, Dict, List

from rich.console import Console
from rich.live import Live
from rich.table import Table

CONSOLE = Console()
COLOR_NORMAL = "white"
COLOR_DEBUG = "grey50"
COLOR_DEBUG2 = "bright_black"
COLOR_INFO = "blue"
COLOR_SUCCESS = "green"
COLOR_WARN = "yellow"
COLOR_ERROR = "red"


class LiveText:
    def __init__(self, text: str = "", refresh_per_second: int = 10):
        self._live = None
        self._running = False
        self._thread = None
        self._text = ""
        self.refresh_per_second = refresh_per_second
        self.info(text)
        self.start()

    @property
    def is_alive(self):
        return self._running

    def _start(self):
        with Live(
            console=CONSOLE, refresh_per_second=self.refresh_per_second
        ) as live:
            self._live = live
            self._live.update(self._text)
            start_time = time.time()
            while self._running:
                elapsed_time = time.time() - start_time
                display_message = (
                    f"[grey0]({elapsed_time: .1f}s)[/grey0] {self._text}"
                )
                self._live.update(display_message)
                time.sleep(1 / self.refresh_per_second)
            self._live.update(self._text)

    def start(self):
        if self._thread and self._thread.is_alive():
            return
        self._running = True
        self._thread = threading.Thread(target=self._start, daemon=True)
        self._thread.start()

    def stop(self):
        self._running = False
        if self._thread:
            self._thread.join()

    def info(self, text: str, end=False):
        self._text = text_color(text, color=COLOR_INFO)
        if end:
            self.stop()

    def warn(self, text: str, end=False):
        self._text = text_color(text, color=COLOR_WARN)
        if end:
            self.stop()

    def success(self, text: str, end=True):
        self._text = text_color(text, color=COLOR_SUCCESS)
        if end:
            self.stop()

    def error(self, text: str, end=True):
        self._text = text_color(text, color=COLOR_ERROR)
        if end:
            self.stop()


def table(
    data: List[Dict[str, Any]],
    title: str = "",
    border: bool = True,
    show_header: bool = True,
    style_cols: Dict[str, str] = None,
    padding: tuple = None,
):
    if not data:
        return
    keys = list(data[0].keys())
    tbl = Table(title=title, show_header=show_header)
    if not border:
        tbl.box = None
        tbl.padding = (0, 2)
    if padding is not None:
        tbl.padding = padding
    for key in keys:
        justify = "right" if isinstance(data[0][key], (int, float)) else "left"
        tbl.add_column(
            key.replace("_", " ").capitalize(),
            justify=justify,
            style=style_cols.get(key, "") if style_cols else None,
        )
    for row in data:
        tbl.add_row(*(str(row.get(key, "")) for key in keys))
    CONSOLE.print(tbl)


def ansi_clean(text: str) -> str:
    return re.sub(r"\x1B\[[0-?]*[ -/]*[@-~]", "", text)


def text_color(text: str, color: str = COLOR_NORMAL) -> str:
    return f"[{color}]{text}[/{color}]"


def _print(text: str, color: str = COLOR_NORMAL) -> None:
    CONSOLE.print(text_color(text, color=color))


def info(text: str) -> None:
    _print(text, COLOR_INFO)


def success(text: str) -> None:
    _print(text, COLOR_SUCCESS)


def warn(text: str) -> None:
    _print(text, COLOR_WARN)


def error(text: str) -> None:
    _print(text, COLOR_ERROR)


def critical(text: str, code: int = 1) -> None:
    error(text)
    sys.exit(code)


def _debug(text: str, color: str) -> None:
    from clifire import application

    app = application.App.current_app
    if not app or app.get_option("verbose"):
        _print(text, color)


def debug(text: str) -> None:
    _debug(text, COLOR_DEBUG)


def debug2(text: str) -> None:
    _debug(f"· {text}", COLOR_DEBUG2)


def var_dump(var) -> None:
    CONSOLE.print(var, highlight=True)



import importlib.util
import os

from clifire import application, out


def load(path):
    if not os.path.exists(path):
        out.debug2("Not exist, skipping")
        return False
    if os.path.isdir(path):
        load_folder(path)
    else:
        load_file(path)
    return True


def load_folder(path):
    for file in os.listdir(path):
        if file.endswith(".py"):
            load_file(os.path.join(path, file))


def load_file(filename):
    out.debug2(f"Loading {os.path.relpath(filename)}")
    module_name = os.path.basename(filename)[:-3]
    spec = importlib.util.spec_from_file_location(module_name, filename)
    module = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(module)
    return module


def main():
    app = application.App(name="CliFire", version="1.0")
    out.debug("Search commands in ./fire folder")
    load(os.path.join(os.getcwd(), "fire"))
    out.debug("Search commands in ./file.py file")
    load(os.path.join(os.getcwd(), "fire.py"))
    app.fire()


if __name__ == "__main__":
    main()



from .. import command, out


class CommandVersion(command.Command):
    _name = "version"
    _help = "Display project version"

    def fire(self):
        out.info(f"{self.app.name} {self.app.version}")



from clifire import out


class Result:
    def __init__(self, code: int = 0, stdout: str = "", stderr: str = ""):
        self.code = code
        self.stdout = self._clean_str(stdout)
        self.stderr = self._clean_str(stderr)
        out.debug2(f"Result: {self}")

    def _clean_str(self, value):
        value = value.decode("utf8") if isinstance(value, bytes) else value
        if value and value.endswith("\n"):
            value = value[:-1]
        return value

    def __bool__(self) -> bool:
        return self.code == 0

    def __str__(self) -> str:
        return self.__repr__()

    def __repr__(self) -> str:
        return str(
            {
                "code": self.code,
                "stdout": self.stdout,
                "stderr": self.stderr,
            }
        )


class ResultOk(Result):
    def __init__(self, stdout: str = ""):
        super().__init__(code=0, stdout=stdout)


class ResultError(Result):
    def __init__(self, stderr: str = "", code: int = 1):
        super().__init__(code=code, stderr=stderr)



from .. import command, out


class CommandHelp(command.Command):
    _name = "help"
    _help = "Show this help"

    command = command.Field(
        pos=1,
        help="Command to show help",
        default=["help"],
    )

    def print(self, txt):
        out._print(txt)

    def get_help(self, cmd):
        lines = (cmd._help or cmd.__doc__ or "").strip().splitlines()
        return lines[0]

    def print_description(self, cmd):
        title = "Description:"
        self.print(f"[bold]{title}[/bold]")
        self.print(f"  {self.get_help(cmd)}")
        self.print("")

    def print_usage(self, cmd):
        title = "Usage:"
        self.print(f"[bold]{title}[/bold]")
        args = [
            f"<{n}>" if cmd._fields[n].is_required else f"[<{n}>]"
            for n in cmd._argument_names
        ]
        cmd_name = cmd._name.replace(".", " ")
        self.print(f'  [cyan]{cmd_name}[/cyan] \[options] {" ".join(args)}')
        self.print("")

    def print_arguments(self, cmd):
        data = []
        for name in cmd._argument_names:
            field = cmd._fields[name]
            data.append({"name": name, "help": field.help})
        if not data:
            return False
        title = "Arguments:"
        self.print(f"[bold]{title}[/bold]")
        out.table(
            data, border=False, show_header=False, style_cols={"name": "cyan"}
        )
        self.print("")

    def print_options(self, cmd):
        options = {}
        for name in cmd._options:
            field = cmd._options[name]
            if isinstance(field, str):
                field = cmd._fields[field]
            name = f"-{name}" if len(name) == 1 else f"--{name}"
            options.setdefault(field, []).append(name)
        if not options:
            return False
        data = []
        for field, names in options.items():
            short = ", ".join(sorted([n for n in names if len(n) == 2]))
            name = ", ".join(sorted([n for n in names if len(n) != 2]))
            data.append({"short": short, "name": name, "help": field.help})
        title = "Options:"
        self.print(f"[bold]{title}[/bold]")
        out.table(
            sorted(data, key=lambda d: d["short"]),
            border=False,
            show_header=False,
            style_cols={"short": "cyan", "name": "cyan"},
        )
        self.print("")

    def print_options_global(self):
        options = {}
        for name in self.app.options:
            field, _value = self.app.options[name]
            if isinstance(field, str):
                field, _value = self.app.options[field]
            name = f"-{name}" if len(name) == 1 else f"--{name}"
            options.setdefault(field, []).append(name)
        if not options:
            return False
        data = []
        for field, names in options.items():
            short = ", ".join(sorted([n for n in names if len(n) == 2]))
            name = ", ".join(sorted([n for n in names if len(n) != 2]))
            data.append({"short": short, "name": name, "help": field.help})
        title = "Global options:"
        self.print(f"[bold]{title}[/bold]")
        out.table(
            sorted(data, key=lambda d: d["short"]),
            border=False,
            show_header=False,
            style_cols={"short": "cyan", "name": "cyan"},
        )
        self.print("")

    def print_commands(self):
        groups = {}
        for cls in self.app.commands.values():
            parts = cls._name.split(".")
            group = " ".join(parts[:-1])
            groups.setdefault(group, []).append(
                {
                    "name": cls._name.replace(".", " "),
                    "help": self.get_help(cls),
                }
            )

        title = "Available Commands:"
        self.print(f"[bold]{title}[/bold]")
        for key in sorted(groups.keys()):
            if key:
                self.print(f"\n  [green]{key}[/green]")
            out.table(
                groups[key],
                border=False,
                show_header=False,
                style_cols={"name": "cyan"},
                padding=(0, 3) if key else None,
            )
        self.print("")

    def fire(self):
        cmd = self.app.get_command(" ".join(self.command))
        self.print_description(cmd)
        self.print_usage(cmd)
        self.print_arguments(cmd)
        self.print_options(cmd)
        self.print_options_global()
        if cmd._name == "help":
            self.print_commands()
